
/**************************************************************
      learn-table
      
      included:
           1. nodes from open-book (default score 0)
           2. nodes from search
      
      
      implementation:
      
          1. save & load text & binary files
          2. quick search learn result for curr. position
          3. past search result
          4. random search move from open-book
          
****************************************************************/




#include "chess.h"


void MakeEpd( char * s );

/*   main node   */
typedef PACKED_STRUCT  // PACKED FLAG FOR GCC
  
LearnNode
{
  HashKey key;              // hash key of position
  struct LearnNode *next;   // next node (list of nodes)

  long  score;              // search result, 0 - default 
  
  short learn;              // not used,  ( count of good play )
  short size;               // size of struct ( + strlen of epd)
  
  char  c;                  // side
  signed char  depth;       // search depth
  char  isFromLib;          // move from open-book ?
  unsigned char
       buf_cntrl_summ;      // summ of epd (for read-write for diff. oper. systems ?!)

  char epd[1]; //0..size    // position string
  
} Node;




/*    hash table, array of lists of learn-node  */

#define SZ (1<<16)  
Node *learn[SZ];     





//---------------------------------------------------------------

void LearnClear(void)
{
  memset(learn,0,sizeof(learn));
}



/* new default node */

Node *NewLNode( char *epd )
{
  Node *p = (Node*) malloc(sizeof(Node) + strlen(epd));

  p->score = p->depth = 0; p->next = 0;
  p->isFromLib = 1;
  p->learn = 16;
  p->key = g.key;
  p->c = g.side;
  p->size = sizeof( Node ) + strlen( epd );
  strcpy( p->epd, epd );
  return p;
}



typedef struct TokType
{
  int hlat[256];
  char *begin, *end;
}
TokType;



void InitToken( TokType * p, char * s, char * del )
{
  int j;
  for ( j = 0; j < 256; j++ ) p->hlat[j] = 0;
  for ( ; * del; del++ ) p->hlat[( unsigned char ) ( * del )] = 1;
  p->begin = p->end = s;
  p->end--;
}

char *GetToken( TokType * p )
{

  if ( * p->begin == '\0' )
    return 0;

  for ( p->begin = p->end + 1; p->hlat[( unsigned char ) ( * p->begin )] == 1 && * p->begin != '\0'; p->begin++ );
  for ( p->end = p->begin; p->hlat[( unsigned char ) ( * p->end )] == 0 && * p->end != '\0'; p->end++ );
  * p->end = '\0';

  if ( * p->begin == '\0' )
    return 0;

  return p->begin;
}


/* 
   read open-book text file 
   sample file:
        //-----------------
        # comment     
        e2e4 e7e5 ....        <---- line 1
        g1f3 g8f6 ....        <---- line 2
        ...
        //-----------------
   
*/
int ReadLibFile( char * fname )
{
  int GetStr( char * s, int n, FILE * f );
  void InsertNodeLearn( void );
  void InitNewGame( void );
  int StrToMove( char * s, Move * ret_mv );

  FILE * f = fopen( fname, "r" );
  if ( f )
  {
    char s[0xFFFF], * strMv;
    Move mv;
    int lines = 0;
    struct Game_Type save = g, new_game;
    TokType t, new_token;
    InitToken( &t, s, " !?#,;\r\n" );

    InitNewGame();
    new_game = g;
    new_token = t;

    while ( GetStr( s, 0xFFFF - 1, f ) )
    {
      lines++;

      t = new_token;  // template for GetToken
      g = new_game;   // new game

      while( (strMv = GetToken( &t )) != NULL )
      {
        if ( StrToMove( strMv, &mv ) && InsertMoveInGame( mv,0 ) )
        {
         
          //  g.side ^= 1; g.xside ^= 1; // set original side for position
          //                           // InsertMoveInGame change side :(( 
            InsertNodeLearn(); 

          //  g.side ^= 1; g.xside ^= 1;

        }else{
          printf( "#error: line %d, lexem %s\n", lines, strMv );
          break; //error
        }
      }
    }
    fclose( f );
    g = save;
    return 1;
  }
  return 0;
}


int GetStr( char * s, int n, FILE * f )
{
  while ( fgets( s, n, f ) )
  {
    char * p = s;
    while ( * p == ' ' ) p++;
    if ( * p != '\0' && * p != '#' && * p != '\n' ) return 1;
  }
  return 0;
}




void InsertNodeLearn( void )
{
  Node * SearchNode( int c, HashKey key );
  void MakeEpd( char epd[] );
  char epd[256];
  Node *node = SearchNode( g.side, g.key );

  MakeEpd( epd );

  if( node != NULL && strcmp( node->epd, epd ) == 0 )
      return; // this position exist

  if(1)      // insert new position
  {
    Node *p = NewLNode( epd );
    p->next = learn[( int )g.key & ( SZ - 1 )];
    learn[( int )g.key & ( SZ - 1 )] = p;
  }

}

Node *SearchNode( int c, HashKey key )
{

  Node * p = learn[( int )key & ( SZ - 1 )];
  while ( p )
  {
    if ( p->key == key && p->c == c )
      return p;
    p = p->next;
  }
  return NULL;
}


/*********  
    position to str, sample:
    
          "r1bqkbnr/ppp1pppp/2n5/3p4/3PP3/8/PPP2PPP/RNBQKBNR w KQkq d3 0 3" 
         
**********/

void MakeEpd( char * s )
{
  const char ch_piece[2] [8] =
  {
    {
      0, 'P', 'N', 'B', 'R', 'Q', 'K', 0
    },
    {
      0, 'p', 'n', 'b', 'r', 'q', 'k', 0
    }
  };


  do
  {
    int x, y, cnt;
    for ( y = 0; y < 8; y++ )
    {
      x = 0;
      while ( x < 8 )
      {
        cnt = 0;
        while ( x < 8 && g.pos[y * 8 + x] == NOPIECE )
        {
          x++;
          cnt++;
        }
        if ( cnt > 0 ) * s++ = '0' + cnt;
        while ( x < 8 && g.pos[y * 8 + x] != NOPIECE )
        {
          * s++ = ch_piece[g.color[y * 8 + x]] [g.pos[y * 8 + x]];
          x++;
        }
      }
      if ( y < 7 ) * s++ = '/';
    }
  }
  while ( 0 );


  *s++ = ' ';
  if ( g.side == WHITE ) *s++ = 'w';
  else
    *s++ = 'b';
  *s++ = ' ';

  do
  {
    int LeftCastlEnable( int c );
    int RightCastlEnable( int c );
    char * p = s;

    if ( LeftCastlEnable( WHITE ) ) * s++ = 'Q';
    if ( RightCastlEnable( WHITE ) ) * s++ = 'K';
    if ( LeftCastlEnable( BLACK ) ) * s++ = 'q';
    if ( RightCastlEnable( BLACK ) ) * s++ = 'k';

    if ( p == s ) * s++ = '-';
  }
  while ( 0 );
  * s++ = ' ';

  do  //0001
  {
    int en_pass_to( Move in_mv, int mv_side );
    extern const char *chb[];
    Move mv = LastMove();
    int sq;
    
    if(sq = en_pass_to( mv, g.xside ))
    {
      *s++ = chb[sq] [0];
      *s++ = chb[sq] [1];
    }
    else
      *s++ = '-';
  }
  while ( 0 );


  for( s--; *s == ' ' || *s == '-'; s-- );
  s++;
  *s = '\0';
}


int pos_ext_key(void){

  int LeftCastlEnable( int c );
  int RightCastlEnable( int c );
  int en_pass_to( Move in_mv, int mv_side );
   
  int sq;
  int d = 0;

  if(sq = en_pass_to(LastMove(), g.xside))  d |= sq;
  
  d |= g.kingSq[WHITE] << 6;
  d |= g.kingSq[BLACK] << 12;
  
  if(g.castl_enable)
  {
    if(LeftCastlEnable(WHITE))               d |= 1<<18;
    if(RightCastlEnable(WHITE))              d |= 1<<19;
    if(LeftCastlEnable(BLACK))               d |= 1<<20;
    if(RightCastlEnable(BLACK))              d |= 1<<21;
  }
  
  return d;
}


int LeftCastlEnable( int c )
{
  if ( c == WHITE )
  {
    if ( g.pos[E1] == KING && g.color[E1] == WHITE && g.pos[A1] == ROOK && g.color[A1] == WHITE
         && ( g.castl_enable & WHITE_LEFT_CASTL_ENABLE ) )
           return 1;
  }
  else
  {
    if ( g.pos[E8] == KING && g.color[E8] == BLACK && g.pos[A8] == ROOK && g.color[A8] == BLACK
         && ( g.castl_enable & BLACK_LEFT_CASTL_ENABLE ) )
           return 1;
  }
  return 0;
}

int RightCastlEnable( int c )
{
  if ( c == WHITE )
  {
    if ( g.pos[E1] == KING && g.color[E1] == WHITE && g.pos[H1] == ROOK && g.color[H1] == WHITE
         && ( g.castl_enable & WHITE_RIGHT_CASTL_ENABLE ) )
           return 1;
  }
  else
  {
    if ( g.pos[E8] == KING && g.color[E8] == BLACK && g.pos[H8] == ROOK && g.color[H8] == BLACK
         && ( g.castl_enable & BLACK_RIGHT_CASTL_ENABLE ) )
           return 1;
  }
  return 0;
}


//0001
int en_pass_to( Move in_mv, int mv_side ){

  int u;

  if(PIECE(in_mv)==PAWN)
   if(abs(FROM(in_mv) - TO(in_mv)) == 16)
   {
       u = TO(in_mv);
       if( COLUMN(u) > 0 )
         if( g.pos[u-1]==PAWN)
          if( g.color[u-1]==mv_side^1)
           return mv_side == WHITE ? (u+8) : (u-8);
       if( COLUMN(u) < 7 )
         if( g.pos[u+1]==PAWN)
          if( g.color[u+1]==mv_side^1)
           return mv_side == WHITE ? (u+8) : (u-8);
   }
  return 0;
}


/************
int PawnFirstExtMove( Move mv, int c )
{
  if ( PIECE( mv ) == PAWN )
  {
    if( ( ROW( FROM( mv ) ) == ( c == WHITE ) ? 6 : 1 ))
     if(  ( ROW( TO( mv ) ) == ( c == WHITE ) ? 4 : 3 ) )
      return 1;
  }
  return 0;
}
*************/



void RndMovesWithLib( int treeLow, int treeHigh)
{
  Node *p;
  int j;
  char epd[1024];

 
  for ( j = treeLow; j <= treeHigh; j++ )
  {

    sort_val[j] = rand()%20;

    MakeMove( tree[j] );
    p = learn[( int )g.key & ( SZ - 1 )];
    while ( p )
    {
      if ( p->key == g.key && 
           p->c == g.side && 
           p->isFromLib 
         )
      {
        MakeEpd( epd );
        if ( strcmp( epd, p->epd ) == 0 )
        {                          //       

          sort_val[j] = rand()%50;  // Up lib moves
          break; 
        }
      }
      p = p->next;
    }//while
    UnMakeMove( tree[j] );
  }
}






/* list moves from open-book */
int ListLibMoves( int treeLow, int treeHigh, Node *v[], Move moves[], int *n )
{
  Node * p;
  int j;
  char epd[1024];

  *n = 0;
  for ( j = treeLow; j <= treeHigh; j++ )
  {
    MakeMove( tree[j] );
    p = learn[( int )g.key & ( SZ - 1 )];
    while ( p )
    {
      if ( p->key == g.key && 
           p->c == g.side && 
           p->isFromLib 
         )
      {
        MakeEpd( epd );
        if ( strcmp( epd, p->epd ) == 0 )
        {
          v[*n] = p;
          moves[*n] = tree[j];
          (*n)++;
          break; 
        }
      }
      p = p->next;
    }//while
    UnMakeMove( tree[j] );
  }
  return (*n) > 0;
}


/*
    
    , 
    - 1000
       -
*/
int extract_strateg_score(int s){

  if(s < -INF+100) return s;
  if(s > INF-100)  return s;  // MATE SCORE
  if(s < 0)
  {
     while(s < VALUE_P/2)
        s += VALUE_P;

  }else if(s > 0){

     while(s > VALUE_P/2)
        s -= VALUE_P;
  }
  return s;
}


/*  get move from open-book */

Move MoveFromLib( int treeLow, int treeHigh )
{
  Node *v[512]; //list of lib. nodes
  Move moves[512];
  static char buf[32];
  int n;
  
  int j;
  Move best_mv = 0;
  int  best_score;
 
  
  
  if ( ListLibMoves( treeLow, treeHigh, v, moves, &n ) )
  {
    int min_score = INF;
    int max_score = -INF;
    const int MIN = -30;
    /*
           ( , 
          - 0)   , 
       .
           -30,     
        -  
    */

    for(j = 0; j < n; j++)
    {
       int s_score = extract_strateg_score(-v[j]->score);

       if(s_score <= MIN)
       {

       }else{
        if(s_score < min_score)
            min_score = s_score;
        if(s_score > max_score)
            max_score = s_score;
       }
    }


    best_score = -INF;
    for(j = 0; j < n; j++)         
    {
      int d,s;

      s = extract_strateg_score(-v[j]->score);

      if(s > MIN)
      {
        
        if(s > 70) s = 70;      // MIN < s <= 70
        
        d = s + min_score + 21;
        //___________________________________
         //??
         //printf("<%s %d> ", MoveToStr(moves[j], buf), d);

        if(d <= 0){
          fprintf(stderr,"# ----> error: MoveFromLib() \n");
        }else{
          int rnd_score = rand()%d;
          if(rnd_score > best_score)
          {
            best_score = rnd_score;
            best_mv = moves[j];
          }
        }
      }
    }
  }

  //??
  //printf("\n");
  return best_mv;
}



unsigned char StrHashKey(char *s){
  unsigned char summ = 0;
  
  while(*s)
  {
    summ += (unsigned char)(*s);
    s++;
  }
  return summ;
}



/* save bin. file */

int SaveLearnTable(char *fname)
{
  FILE * f = fopen( fname, "wb" );
  if ( f )
  {
    int j;
    for ( j = 0; j < SZ; j++ )
    {
      Node *p = learn[j];
      while( p )
      {
        p->buf_cntrl_summ = StrHashKey(p->epd);
        fwrite( p, 1, p->size, f );
        p = p->next;
      }
    }

    fclose( f );
    return 1;
  }
  return 0;
}


/*  load bin. learn file  */

int LoadLearnTable( char * fname )
{
  FILE * f = fopen( fname, "rb" );
  if ( f )
  {
    Node * p;
    char * buf;
    int add_size, tmp;
    int f_size;
    int readCnt = 0;
    
    
    //get file-size
    fseek( f, 0, SEEK_END );
    f_size = ftell( f );
    fseek( f, 0, SEEK_SET );
    
    
    //get memory
    buf = (char*)malloc( f_size );
    if ( buf == 0 )
      goto error;
      
      
    //read file 
    //0001
    do
    {
      tmp = fread(buf, 1, sizeof(Node), f);
      if(tmp == 0) goto success;
      if(tmp != sizeof(Node))
        goto error;
      
        
      p = (Node *) &buf[0];               // 0001
      add_size = p->size - sizeof(Node);
      if(add_size <= 0) goto error;
      
      
      if(fread(buf + sizeof(Node), 1, add_size, f) != (unsigned)add_size)
        goto error;
        
        
      if(p->buf_cntrl_summ != StrHashKey(p->epd))
        goto error;
        
        
      buf += p->size;
      readCnt += p->size;
      p->next = learn[(int)p->key & (SZ - 1)];
      learn[(int)p->key & (SZ - 1)] = p;

    }
    while ( 1 );


  success:
    fclose( f );
    return 1;
  error:
    fclose( f );
    LearnClear();
  }
  return 0;


}


/* insert curr. position */

void SaveSearchResult( int score, int depth )
{
  Node * p = learn[(int)g.key & (SZ - 1)];
  char epd[256];

  MakeEpd( epd );
                                //search exist node
  while( p )
  {
    if( p->key == g.key && p->c == g.side && strcmp( epd, p->epd ) == 0 )
    {
      p->depth = depth;
      p->score = score;
      return;
    }
    p = p->next;
  }
                                // insert new node
  p = NewLNode( epd );
  p->next = learn[( int )g.key & ( SZ - 1 )];
  learn[( int )g.key & ( SZ - 1 )] = p;

  p->score = score;
  p->learn = 0;
  p->c = g.side;
  p->depth = depth;
  p->isFromLib = 0;
}


/* search current position in learn-table */

int LearnLook( int depth, int alpha, int beta, int *ret_score )
{
  Node * p = learn[( int )g.key & ( SZ - 1 )];
  char epd[256];

  while ( p )              
  {
    if ( p->key == g.key && p->c == g.side )
    {
      MakeEpd( epd );
      if ( strcmp( epd, p->epd ) == 0 )
      {

         // ??
         // extern int ply;
         // printf("-----------> learn look side=%d, ply=%d \n", g.side, ply);
         // fflush(stdout);

        if ( p->depth >= depth )
        {
          *ret_score = p->score;
          return 1;
        }
        else if ( p->isFromLib && p->depth==0 ) // from open-book and not searched
        {
          const int margin = 16;   // default score:  -margin..0..margin
          
          if ( 0 + margin <= alpha )
          {
            *ret_score = alpha;
            return 1;
          }else if ( 0 - margin >= beta )
          {
            *ret_score = beta;
            return 1;
          }

        }
        return 0;
      }
    }
    p = p->next;
  }

  return 0;
}

