#include "chess.h"

  extern int is_rnd_search;

  extern int property_load(char fname[]); // in property.c
  extern char* property_get(char name[]);
  extern void property_add(char name[], char value[]);
  extern void InitNewGame( void );
  extern int prog_is_run(void);
  extern int StrToMove( char *s, Move *ret_mv );
  extern int SearchMove( Move *ret );
  extern int InitGame( char *epd );
  extern int ReadLibFile( char *fname );
  extern int SaveLearnTable( char *fname );
  extern int LoadLearnTable( char *fname );
  extern int search_score; // in search.c
         int Resigns(int cnt, int low_bound);
  extern int load_play_style(char s[]);
  extern char* print_play_str(void);

  int move_machine(void); 


//------------------------------------

//var. for Play Loop

extern char  cfg_user_def_fname[1024];
extern char  prog_dir[1024];
char         game_fname[1024];


int is_echo_board = 0;
int is_xboard = 0;
int stop = 0;
int prev_inst = 0;



//-------------------------------------


char *get_board(void){

  void MakeEpd( char * s );
  static char buf[256];
         
  MakeEpd( buf );
  return buf;
}

unsigned file_size(char fname[]){
  FILE *f = fopen(fname,"rb");
  unsigned f_size;

  if(f==0){
      fprintf(stderr,"------> error: file_size() - file %s not open\n", fname);
      return 0; 
  }
  fseek( f, 0, SEEK_END );
  f_size = ftell( f );
  fclose(f);

  return f_size;

}


static char old_print_fen[1024];

int print_board(){

  
  char *fen = get_board();
  
  if(strcmp(fen,old_print_fen) != 0)
  {
    strcpy(old_print_fen,fen);
    printf("position %s\n", fen);
  
  }
  return 0;
}




void print_moves(char fname[]){

	        FILE *f = stdout;

		if(fname)  f = fopen(fname,"w");
	

		if(f)
		{  char * MoveToStr( Move mv, char * str );
		   static char s[32];
		   int j;
                   if(f == stdout)
 		      fprintf(f, "moves ");				  
                   else{
                      char *comment = strtok(0, " \n");
                      if(comment==0) comment = "player - chess";
                      fprintf(f, "[Event  \"%s\"] \n", comment);
                   }
		   for(j = 0; j < g.game_cnt; j++)
		   {
			   fprintf(f,"%s ", MoveToStr(g.game_list[j],s));
			   
		   }
                   fprintf(f, "\n");
                   fflush(f);

		   if(f != stdout) fclose(f);

  	        }else
		   fprintf(stderr,"error open file\n");


}


char *ConCat(char *s1, char *s2)
{

  static char buf[8][1024];
  static int  cnt = -1;
  char *s;

  if(++cnt > 7) cnt = 0;
  s = buf[cnt];
  
  strcpy( s, s1 );
  strcat( s, s2 );
  
  return s;

}




/* usage:  use_moves(g.game_cnt) */

int  use_moves(int cnt){
   // { 0, 1,1,  2,2, 3,3, 
   // 0 --> 0
   // (cnt-1)/2 + 1

   if(cnt==0) return 0;
   else return (cnt-1)/2 + 1;

}



void show_board(void){

   const char *chp[2] = 
   {
     " PNBRQK",
     " pnbrqk" 
   };
   const int black_sq[64] = 
   {
     0, 1, 0, 1, 0, 1, 0, 1,
     1, 0, 1, 0, 1, 0, 1, 0,
     0, 1, 0, 1, 0, 1, 0, 1,
     1, 0, 1, 0, 1, 0, 1, 0,
     0, 1, 0, 1, 0, 1, 0, 1,
     1, 0, 1, 0, 1, 0, 1, 0,
     0, 1, 0, 1, 0, 1, 0, 1,
     1, 0, 1, 0, 1, 0, 1, 0
   };
   const char rank[8] = {'8', '7', '6', '5', '4', '3', '2', '1'};
   const char file[8] = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'};
   
   
   int x,y,sq,p;
   
   for(y = 0; y < 8; y++)
   {
   
     printf("%c  ", rank[y]);
     
     for(x = 0; x < 8; x++)
     {
        sq = y*8 + x;
        p = g.pos[sq];
        if(p==NOPIECE)
        {
           printf("%c ", black_sq[sq] ? ' ':'-'); 
        
        }else{
           printf("%c ", chp[ g.color[sq]][p]);        
        
        }
     }
     printf("\n");
   }
   
   printf("\n   ");
   for(x = 0; x < 8; x++)
     printf("%c ", file[x]);
   printf("\n");
     
     
   printf("moves: %d\n",  use_moves(g.game_cnt)); 
   printf("side:  %s\n",  g.side==WHITE?"white":"black");
}







void PlayWinBoard( void )
{
  #define LEX_IS(s)  (strcmp(s,lex)==0)
  
  char s[1024];
  Move mv,find_mv;
  
  
  
  prev_inst = prog_is_run();
//0005

  setbuf( stdout, NULL );
  setbuf( stdin, NULL );
  
 

                                    //load data & print status
  
  printf( "# Chess, build %s \n",  __DATE__);
  
  
  
  
  switch( prev_inst )
  {
     case 0:   // learn write enable
      //...
     break;
     case 1:
       printf("# warning: learn file write disable \n");
     break;
     case 2:
       printf("# warning: obender(2): learn file write disable \n");
     break;
  }
  
  

  
  
  
  
  
  // load *.cfg file
  
  
  if(property_load( ConCat(prog_dir, PROPERTY_DEFAULT_NAME)))
  {
    
  }else{
    fprintf(stderr, "conf.file not found\n");
    exit(1);
  }    


  InitAllVar();
  InitNewGame();


 
  if(LoadLearnTable( ConCat(prog_dir, LEARN_DEFAULT_NAME) ))
  {
  
  }else
  {
    extern void LearnClear(void);
    LearnClear();
    remove(ConCat(prog_dir, LEARN_DEFAULT_NAME));
    
    fprintf(stderr, "#  learn file notfound, load book-file \n");  
    
    
    if( ReadLibFile( ConCat(prog_dir, BOOK_DEFAULT_NAME))==0)
    {
      fprintf(stderr,"book file notfound  \n");
      exit(1);
    }
  }
    
  
  strcpy(game_fname,  ConCat(prog_dir, GAME_DEFAULT_NAME)  );  
  
  
  
  if( strcmp(property_get("random search?!"),"enable")==0)
      is_rnd_search = 1;
  else
      is_rnd_search = 0;


  //extern time_t limit_time;
  limit_time = 100*atoi(property_get("time search default seconds"));
  if(limit_time==0)
    printf("#warning: limit time error in cfg file \n");
  if(limit_time < 7*100) limit_time = 7*100;
  if(limit_time > 100*100) limit_time = 100*100; 


  printf("# load play style: \n");
  if(load_play_style(property_get("play style"))==0)
      printf("# error \n");
  else{
      printf("# %s \n", print_play_str());
      printf("# ok \n");
  }   
  
  
  
  
  printf( "# <help> - help \n");
  printf( "# move: ? \n" );
  
  
  
  //0005
  
  // main loop
  
  while ( fgets( s, sizeof( s ) - 1, stdin ) )
  {
    char *lex = strtok( s, " \n" ), *pen;
    
    int put_move = 0;
    
    if ( lex )
    {
      if(strcmp(lex,"getboard")==0)
        strcpy(old_print_fen,"");    // reset as default
   
    
      if(is_echo_board)
        print_board();
    
      if ( LEX_IS( "quit" ) ) break;
      //0005
      else if( strcmp(s,"echoboard")==0 ) is_echo_board = 1;
      else if( strcmp(s,"getmoves")==0 )
      {
         char *fname = strtok(0, " \n");
        
         print_moves(fname);
      
      }
      else if( strstr(s,"help")   )
      {
        printf("#  move                                     \n");
        printf("#   <from><to>[promote_piece]               \n");
        printf("#    example:                                \n");
        printf("#          e2e4,  a6a7q                     \n");
        printf("#  <quit>            - exit and save learn. \n");
        printf("#  <undo>            - go to back           \n");
        printf("#  <redo>            - ---                  \n");
        printf("#  <new>             - new game             \n");
        printf("#  <force>           - for input moves      \n");
        printf("#  <st> <second>     - time per. move       \n");
        printf("#  <go>              - get move             \n");
        printf("#  <setboard>  <fen_pos> - put position     \n");
        printf("#  <getboard>            - get position     \n");
        printf("#  <pos>             - show position        \n");
        printf("#  <getmoves>        - print all moves      \n");
        printf("#  <style> [level]   - knowledge1, knowledge2, knowledge3 \n");
        printf("#  ? <rnd> [on] [off]  - new experement. search \n");
        printf("#  <save> [file], <load> [file] - save(load) game in binary format \n");
        printf("#  examples:                         \n");
        printf("#        xboard -fcp ./obender              \n");
        printf("#        winboard.exe -fd -fcp obender.exe          \n");
        printf("#        echoboard.exe C:\\Games\\Obender\\obender.exe chess \n");
        printf("#  search  output format:                   \n");
        printf("#        depth  score  time(1/100sec.)  pos/sec   pv-line         \n");
        printf("#  about:                                   \n");
        printf("#       %s  \n",  PROG_AUTHOR);
        printf("#       %s  \n",  PROG_SITE);
        printf("#       %s  \n",  PROG_MAIL);
      }      
      else if ( LEX_IS( "xboard" ) )
      {
          is_xboard = 1;
          printf( "feature setboard=1\n" );
          printf( "feature done=1\n" );
          printf( "feature sigint=0\n" ); // only for Linux
          printf( "feature sigterm=0\n" );  
      }
      else if ( LEX_IS( "st" ) )
      {
        extern time_t limit_time;
        //static int cnt_warning = 0;
        char *s = strtok(0, " \n" );

        if(s==0 || atof(s)==0)
            fprintf(stderr,"----> error: expected <st second's>\n");
        else
      {

        limit_time = atof(s) * 100; // $  st <seconds>
        
        //??
        //printf("---> %s (%d) \n", s, limit_time);


        if(limit_time < 4*100)
        {
          
         // if(++cnt_warning == 0)
          fprintf(stderr, "# warning:  time = 4 sec \n");
          limit_time = 4*100;
        }
        if(limit_time > 100 * 100)
        {
          //if(++cnt_warning == 0)
          fprintf(stderr, "# warning:  time = 100 sec \n");  
          limit_time = 100 * 100;
        }
     } 
      }
      
      else if( LEX_IS( "time" ) )  //  time <seconds * 100>
      {
        extern time_t limit_time; // in timer.c
        static int cnt_warning = 0;
        char *s = strtok( 0, " \n" );
        time_t tmp;
        static time_t old_time;
        /**********
        if(g.side==WHITE)
        {
           limit_time = atoi(s)      
                          / 
                       (40 - (g.game_cnt/2)%40);
                       
        }else{                               // thanks JA
           limit_time = atoi(s) 
                          / 
                       (39 - (g.game_cnt/2)%39); // ??
        }
        ************/
        
        
        /************
        if(g.side == WHITE)
        {
          limit_time = atoi(s) / (40 - (g.game_cnt/2)%40);

        }else{
          limit_time = atoi(s) / (40 - (g.game_cnt/2+1)%40);
        }
        *************/
        
        
        // 0001
        
        //  !!! warning:  comment this block if not work and
        //                use other algoritm
         
        
        tmp = atoi(s) / (40 - use_moves(g.game_cnt)%40);
        
        
        old_time = 750;  
        
        if( tmp > old_time * 2)
        {
           SWAP(tmp,old_time);
        }else
           old_time = tmp;
           
        
        
        if(tmp < 7*100)
        {
          if(++cnt_warning == 0)
            fprintf(stderr, "# warning:  time = 7 sec \n");
          tmp = 7*100;
        }
        if(tmp > 30 * 100)
        {
          if(++cnt_warning == 0)
            fprintf(stderr, "# warning:  time = 30 sec \n");  
          tmp = 30 * 100;
        }
           
        limit_time = tmp;   



      }
      
      
      
      else if ( LEX_IS( "force" ) )
      {

        stop = 1;

      } else if(LEX_IS("rnd")){
         extern int is_rnd_search;
         char *s = strtok( 0, " \n" );

          if(s && strcmp(s,"on")==0)
              is_rnd_search = 1;
          else if(s && strcmp(s,"off")==0)
              is_rnd_search = 0;
          else fprintf(stderr,"error: expected arg, [on] or [off]\n");

     

      }
      else if ( LEX_IS( "undo" ) )
      {
        if ( g.game_cnt > 0 )
        {
          Move mv = g.game_list[g.game_cnt - 1];
          UnMakeMove( mv );
          g.game_list[g.game_cnt - 1] = mv;
          g.game_cnt--;
          ply = 0;
        }
      }
      else if ( LEX_IS( "up" ) )
      {
        while( g.game_cnt > 0 )
        {
          Move mv = g.game_list[g.game_cnt - 1];
          UnMakeMove( mv );
          g.game_list[g.game_cnt - 1] = mv;
          g.game_cnt--;
          ply = 0;
        }
      }
      
      else if ( LEX_IS( "redo" ) )
      {
        if ( g.game_cnt < g.game_max )
        {
          MakeMove( g.game_list[g.game_cnt] );
          g.game_cnt++;
          ply = 0;

        }
      }
      else if ( LEX_IS( "end" ) )
      {
        while ( g.game_cnt < g.game_max )
        {
          MakeMove( g.game_list[g.game_cnt] );
          g.game_cnt++;
          ply = 0;
        }
      }
      else if ( LEX_IS( "new" ) )
      {
        InitNewGame();
        stop = 0;

      }
      else if ( LEX_IS( "protover" ) )
      {
        printf( "feature done=1\n" );
      }
      else if( LEX_IS("style")){
        int load_play_style(char s[]);
        char* print_play_str(void);

        char *p = strtok(NULL," \n");
        if(p==0)
            printf("# %s \n", print_play_str());
        else if(load_play_style(p)==0 ) 
           fprintf(stderr,"# unknow style\n");

      }
      else if ( LEX_IS( "deb" ) )
      {
         char *p = strtok(NULL," \n"); 
		 char *pos = "";
         if(p)

		  switch(atoi(p))
		  {
		     case 1:  pos = "r4rk1/4Rp2/6p1/1p1p1bNp/p1qP3P/2P1Q3/P4PP1/4R1K1 b - - 0 0";
				      break;
		     case 2:  pos = "r1bq1r1k/ppp2pp1/2nbp2p/6NQ/3P4/2PB4/P1P2PPP/R1B2RK1 b - - 0 12";
				      break;
		     case 3:  pos = "2r2b1r/p1qk1ppp/2p1pn2/1B6/Pp6/4PP2/1PQB1P1P/2R2RK1 w - - 0 16 "; 
				      break;
		     case 4:  pos = "r5k1/7p/1Rp2rp1/2Pp4/P2Qp3/4PqP1/3R1P1P/6K1 w - - 0 38";
				      break;
			 case 5: pos = "rnbqkbnr/ppp1pppp/8/3pP3/8/8/PPPP1PPP/RNBQKBNR b KQkq - 0 2";
				      break;
			 case 6: pos = "rnbqkbnr/ppp1pppp/8/8/2pP4/8/PP2PPPP/RNBQKBNR w KQkq - 0 3";
				      break;
			 case 7: pos = "r1b1kb1r/ppp1n3/3q4/4pp2/4Q1p1/3B4/PPPP1PP1/R1B1NRK1 w kq - 0 15";
				 break;
			 case 8: pos = "4b3/2R3B1/5P2/pk5p/2p2r1P/P1K5/1P6/8 w - - 0 63";
				  break;
			 case 9: pos = "r3kb1r/ppp1pppp/2n2n2/q7/2BP2b1/2N5/PPP1NPPP/R1BQK2R w KQkq - 5 7 ";
				  break;
		 	 case 10: pos = "r3kb1r/ppp1pppp/2n2n2/q4b2/2BP4/2N2P2/PPP1N1PP/R1BQK2R w KQkq - 1 8 ";
				 break;
			 case 11: pos = "3k4/8/2K1Q3/8/8/8/8/8 w - - 0 1";
			          // white mate with 1 move
//x001			          
				 break;
		 }    
         
		 if(InitGame(pos)==0) fprintf(stderr,"#error\n");
      }

      else if ( LEX_IS( "setboard" ) &&  (pen = strtok(NULL,"\n")) != NULL
           && InitGame( pen ) )
      {
          
          
      }
      
 //0005     
      else if( LEX_IS("getboard") )
      {
        print_board();
      
      }
      
      
      
//x001      
      else if ( StrToMove( lex, &mv ) && InsertMoveInGame( mv, UNKNOW_SEARCH_REZULT ) )
      {
         
        //0005
        if(is_echo_board)
           print_board();
           
         //x002
        if ( !stop )
           move_machine();

      }
      
      
      else if ( LEX_IS( "go" ))
      {
        //x001
        stop = 0;
        
        move_machine();
      }
      
      
      
      
      else if( LEX_IS("pos")){
        //0001
        show_board();
      }
      
      
      
      
      
      else if(LEX_IS("load"))  //0001
      {
        struct Game_Type tmp;
        FILE *f;
        char *fname;
        int ok;
        char *s = strtok(0, " \n" );

        if(s && *s) fname = s;
        else fname = game_fname;
                
        ok = 0;
        if(f = fopen(fname,"rb"))
        {
         if( fread(&tmp, sizeof(struct Game_Type), 1, f)== 1)
          if(tmp.id == 11)
          {
            g = tmp; 
            ok = 1;
            strcpy(game_fname, fname);
            set_rep();
          }
         fclose(f);
        }  
        if(!ok) fprintf(stderr, "# error open file %s \n", fname);   
      }
      
      
      else if(LEX_IS("save"))
      {

        FILE *f;
        int ok;
        char *fname;
        char *s = strtok(0, " \n" );

        if(s && *s) fname = s;  // new file name
        else fname = game_fname;
        
        ok = 0;
        if(f = fopen(fname,"wb"))
        {
         g.id = 11;
         if( fwrite(&g, sizeof(struct Game_Type), 1, f)== 1)
         {
            ok = 1;
            strcpy(game_fname, fname);
         }
         fclose(f);
        }  
        if(!ok) fprintf(stderr, "# error write file %s \n", fname);   


                 
      }
      
      
      
      else if(LEX_IS("score")){
        void InitEvaluate();
        
        InitEvaluate();
        
        Evaluate(-INF,INF);
        //print_score_tbl();
      
      }
      

      else if(LEX_IS("getmoves")){
        int j;  
          
        for(j = 0; j < g.game_cnt; j++)
          printf("%s ", MoveToStr(g.game_list[j], 0));  
          
        printf("\n");          
      }


      
      else if(LEX_IS("test")){              // 0001
      
        /************
         int j;
                  
         for(j = 0; j < 82*2; j++)
         {
           //if(j==79 || j==80*2-1)
           if(use_moves(j)%40==0)
             printf("   ---->");
           printf("(%d)%d ", j, 40-use_moves(j)%40);
           
         }

         printf("\n");
       **************/
         
         int Repetition(void);
         
         printf("Repetition - %d \n", Repetition() );
       
       
      }      
      
      else{
        fprintf( stderr, "# unknow: %s\n", lex );
      }

//0005
        if(is_echo_board)  // && !put_move)   // not owerride move
          print_board();


    } //if lex
  } //while

if( prev_inst==0)
  if( strcmp(property_get("learn file write"),"disable") != 0 )
  {
    char tmp_file_name[64]; 
    sprintf(tmp_file_name,"%s%d.lrn", "obender_tmp_lrn", rand()%1000);


    if( SaveLearnTable( ConCat(prog_dir, tmp_file_name) )==0)
      fprintf(stderr, "error write learn file\n");
    else{
    //     
    //      
    // ,     
    //      
      
      int sz1,sz2;

   
      sz1 = file_size( ConCat(prog_dir, tmp_file_name) );
      sz2 = file_size( ConCat(prog_dir, LEARN_DEFAULT_NAME) );

      if(sz1 < sz2)  
      {
         fprintf(stderr, "---->  error write learn file(1), size new file less old file \n");
      }else{
        remove(ConCat(prog_dir, LEARN_DEFAULT_NAME));
        if(rename(ConCat(prog_dir, tmp_file_name),
            ConCat(prog_dir, LEARN_DEFAULT_NAME)) != 0)
            fprintf(stderr,"----> error write learn file(2), ?rename() \n");
        remove(ConCat(prog_dir, tmp_file_name));
        printf("write learn file %s %d bytes \n", ConCat(prog_dir, LEARN_DEFAULT_NAME),
               file_size( ConCat(prog_dir, LEARN_DEFAULT_NAME)) );
      }
    }
  }
   
   
}//endp PlayWinBoard

//----------------------------------
/***

  Computer resigns

            score history diagram 
   
  | * *
  |      *  
  |        *
  |          *  + 2 PLY
  |            * <-----  resign point ?!
  |
  |
  |------------------------>s
   
***/




int Resigns(int cnt, int low_bound)
{

 
  int *v = &g.search_result[0];
  
  
  int j          = cnt;
  int cnt_resign = 0;
  
  loop:
      if(j >= 2)
       if(v[j] < v[j-2])
        if(v[j-2] <= low_bound)
        {
          j -= 2;
          cnt_resign++;
          goto loop;
        }
  

  return cnt_resign;

  /***********  
  if(cnt >= 8)
   if(v[cnt] < low_bound)
     if(v[cnt] < v[cnt-2])
       if(v[cnt-2] < v[cnt-4])
        if(v[cnt-4] < v[cnt-6])
          if(v[cnt-6] < low_bound)
          {
              if(v[cnt-6] < v[cnt-8])
               if(v[cnt-8] < low_bound)
                  return 2; 
                
              return 1; // <------ resign, 3 ( 4 ?) point(s) 
          
          }
  return 0;        
  ******************/  

}//endp Risigns


int move_machine(void){
  Move find_mv;
  extern int search_score;
 
          if ( SearchMove( &find_mv ) && InsertMoveInGame( find_mv, search_score ) )
          {
            char str_move[64];
            char *MoveToStr( Move mv, char * str );
            
            //0005
           if(is_echo_board)              // PRINT POS
             print_board();
           else if((is_echo_board|is_xboard)==0)      
             show_board();
            
            
            
            printf( "move %s\n", MoveToStr( find_mv, str_move ) );
            //put_move = 1;
            if( Resigns( g.game_cnt-1, -VALUE_Q)==4)
              printf( "computer resigns\n" );
            else {
            //x002
              extern int Check();

              LegalMoveList();
              if(treeCnt[1]-treeCnt[0]==0)
                 printf("the end\n");
              else if(is_xboard==0)
                     if(Check())
                         printf("check\n");
            }  
            
            return 1;
          }else{ 
          
            //if(g.game_cnt >= 2)
            // if(g.search_result[g.game_cnt-2] < -INF-100)
            //  printf("congratulations!\n");
            printf( "the end\n" );
            return 0;
          }
}




